//-----------------------------------------------------------------------------
//
//  File:       BRMK3.c
//  Author:     Bruce Sherry
//  Date:       7/26/2006 6:44PM
//  Purpose:    Program for BrassRat Regulator Mark3
//  needed
//  Software:   AVR-GCC to compile
//  needed
//  Hardware:   AMega88
//  Revision:   1.01
//
//  Modification History:
// 1.00 ->      Converted from RRMK3-2.c
// 1.00 -> 1.01 Converted from RRMK3phev.c BS 7/26/2006 6:44PM

//-----------------------------------------------------------------------------

#define SIGNON_STRING "--Brassrat Regulator\n"
#define VERSION_STRING "--V1.01 Unit: "

#define USETIMER2
#define LONG_RESPONSE
#define ALLOW_SHUNT

#define DEBUG

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <string.h>
#include <stdarg.h> // sheer added to support variable args to uart_putstring

#include <stdio.h>
#include <avr/eeprom.h>
#include "BRMK3.h"
#include <avr/pgmspace.h>
#include <ctype.h>

/******************************************************************************
*
* Serial Protocol
*
* The first two bytes are the device address in ascii decimal, 1 to 99, with 99
* addressing all boards, which is only valid for setting min and max voltage.
*
* commands can be abreviated by typing to the unique character, and ending
* with a period '.'.
* 01voltage         Unit 1 voltage query: returns voltage in ascii
* 02v.              Unit 2 voltage query: returns voltage in ascii
* 03sethigh 15.55   Set Unit 03 regulation voltage to 15.55V
* 99seth. 14.50     Set all units regulation voltage to 14.50V
* 04setlow 10.00    Set Unit 4 undervoltage warning level to 10.00V
* 05minvolts        Read unit 5 minimum voltage recorded
* 06maxvolts        Read unit 6 maximum voltage recorded
* 07hstclear        Clear unit 7 recorded min and max voltages
* 99h.              Clear all units min and max history
* 08status          Read status of unit 08. Returns a hexadecimal value as the
*                   sum of:
*                       Disabled        0x80
*                       Shunting        0x40
*                       Under Voltage   0x20
*                       Over Voltage    0x10
*                       Has Been Low    0x08
*                       Over Temp       0x04
*                       Normal Oper.    0x00
* 09eepromwr xxx yy Write value 'yy' to eeprom location 'xxx' on unit 9
* 10gethighv        Read regulation voltage on unit 10
* 11readlowv        Read low voltage warning limit on unit 11
* 13temperat        Read unit temperature
* 12xtrntemp        Read external temperature on unit 12
* 14disable         Disable Shunt on board 14
* 15enable          Enable shunt on board 15            
* 16v2volts         Read Voltage 2 on board 16
* 17v3volts         Read Voltage 3 on board 17          
*
******************************************************************************/

// AVR306: Using the AVR UART in C
// Routines for interrupt controlled USART
// Last modified: 02-06-01
// Modified by: AR

/* UART Buffer Defines */
#define BAUD_CONSTANT 10            /* used for broadcast mode sleep */

#define USART_BUFFER_SIZE 64     /* 2,4,8,16,32,64,128 or 256 bytes */

uint8_t * command_pointer;

typedef struct fifo
{
    uint8_t get;                        // get pointer
    uint8_t put;                        // put pointer  
    uint8_t buffer[USART_BUFFER_SIZE];  // fifo buffer
} fifo;

        fifo    rxchars;                // fifo for the receive
        fifo    txchars;                // fifo for the transmit

/* Interrupt handlers */

//void SIG_USART_RECV() __attribute__ ( ( signal ) );
#ifdef USETIMER1
    void SIG_OVERFLOW1() __attribute__ ( ( signal ) );
#else
    void SIG_OVERFLOW2() __attribute__ ( ( signal ) );
#endif
/* Function Prototypes */
void USART_Init( uint16_t baudrate );
uint8_t uart_getchar( void );
uint8_t CmdReady( void );
uint8_t DataInBuffer(const fifo *bf);
void uart_putchar( uint8_t data );
void uart_send();
void uart_putstring(char * fmt, ...);
uint16_t ATOD_read(uint8_t channel);
uint8_t convert_digit(void);
uint8_t convert_voltage(uint8_t* output, uint16_t value, uint16_t cal, int16_t off);
void convert_temperature(uint8_t* output, uint16_t value, uint16_t cal, int16_t off);
void sleep(uint16_t pTicks);
void evaluate_command(uint8_t *cmd_str);
//void get_command(uint8_t * cmd_str, uint8_t length);
uint8_t *get_string();
uint8_t get_address(uint8_t ch1, uint8_t ch2);
uint16_t filter(uint16_t value, uint16_t volatile * last);

/* Prototypes */
void USART0_Init( unsigned int baudrate );
uint8_t USART0_Receive( void );
void USART0_Transmit( uint8_t data );

#define PGM_P   const prog_char *  
 

#define OSCILLATOR 8000000
//#define BAUDRATEDIV 103       /* 4800 */
//#define BAUDRATEDIV 207       /* 2400 */
#define BAUDRATEDIV 51       /* 9600 */

#define TCCR1_DIV 1
#ifdef USETIMER1
#define TCCR1_MODE 14
#else
#define TCCR1_MODE 0
#endif

#define VOLT_OFFSET -0
#define TEN_VOLTS 408L
#define NORM_VOLTS 500L
#define HIGH_VOLTS 605L
#define LOW_VOLTS 409L
#define VOLTSCALNUM 626
#define VOLTS_LEN   6
#define TEMP_OFFSET -0
#define HUNDRED_DEGREES 205L
#define TEMPCALNUM 125
#define TEMP_HOT 307
#define ADCPSCALE 6
#define ADCFAKE (0<<ADEN) | (0<<ADSC)
#define ADCSTART ((1<<ADEN) | (1<<ADSC) | ADCPSCALE)
#define ADCREF 0
#define MAXCMDLEN 20
#define BLINKDELAY 64
#define FILTER_CONST 4

#define ANALOG_PINS 0x3F

#define DISABLED_STATUS 0x80
#define SHUNTING_STATUS 0x40
#define UNDERVOLT_STATUS 0x20
#define OVERVOLT_STATUS 0x10
#define BEENLOW_STATUS 0x08
#define OVERTEMP_STATUS 0x04

#define VOLTAGE_OFFSET_ADDR 0
#define VOLTAGE_CAL_ADDR 2
#define TEMPERATURE_OFFSET_ADDR 4
#define TEMPERATURE_CAL_ADDR 6
#define V2_OFFSET_ADDR 8
#define V2_CAL_ADDR 0x0A
#define V3_OFFSET_ADDR 0x0C
#define V3_CAL_ADDR 0x0E
#define LOW_VOLTS_ADDR 0x10
#define HIGH_VOLTS_ADDR 0x12
#define TEMP_HOT_ADDR 0x14
#define SERIAL_NUMBER_ADDR 0x16
#define BUS_UNIT_ADDR 0x1A
#define OSCCAL_BYTE_ADDR 0x1B

/***********************************************************************
*      DIV        ticks
*       7           .125 8 seconds
*       6           .5  2 seconds
*       5           1
*       4           2   500mS
*       3           4   250mS
*       2          16    15.66mS
*       1         128     7.81mS
*
***********************************************************************/

#define TICKS_PER_SECOND 256
#define TCCR2_DIV 1
#define TCCR2_MODE 7
#define TCCR2_FREQ 32768
//#define OCR2_COUNT (TCCR2_FREQ / TICKS_PER_SECOND) - 1
#define OCR2_COUNT 127
#define ICR1_COUNT (OSCILLATOR / TICKS_PER_SECOND) - 1
#define TIME_TILL_SLEEP 15
#define TICKS_TILL_SLEEP TIME_TILL_SLEEP * 60 * TICKS_PER_SECOND
// calibrate once per second for now.
#define CALIBRATION_MASK ((1L << 10) - 1)

#define MAX_OSC_COUNT 4985
#define MIN_OSC_COUNT 4780

/* These appear in memory in reverse order */
const   uint8_t ENDL[]  PROGMEM = {0x80};
const   uint8_t EXTP[]  PROGMEM = {'x','t','r','n','t','e','m',('p' | 0x80)};
const   uint8_t RDLO[]  PROGMEM = {'r','e','a','d','l','o','w',('v' | 0x80)};
const   uint8_t GETH[]  PROGMEM = {'g','e','t','h','i','g','h',('v' | 0x80)};
const   uint8_t EEPV[]  PROGMEM = {'e','e','p','r','o','m','w',('r' | 0x80)};
const   uint8_t ENAB[]  PROGMEM = {'e','n','a','b','l',('e' | 0x80), 0, 0};
const   uint8_t DISA[]  PROGMEM = {'d','i','s','a','b','l',('e' | 0x80), 0};
const   uint8_t HCLR[]  PROGMEM = {'h','s','t','c','l','e','a',('r' | 0x80)};
const   uint8_t MAXV[]  PROGMEM = {'m','a','x','v','o','l','t',('s' | 0x80)};
const   uint8_t MINV[]  PROGMEM = {'m','i','n','v','o','l','t',('s' | 0x80)};
const   uint8_t SETL[]  PROGMEM = {'s','e','t','l','o',('w' | 0x80), 0, 0};
const   uint8_t SETH[]  PROGMEM = {'s','e','t','h','i','g',('h' | 0x80), 0};
const   uint8_t V3VO[]  PROGMEM = {'v','3','v','o','l','t',('s' | 0x80), 0};
const   uint8_t V2VO[]  PROGMEM = {'v','2','v','o','l','t',('s' | 0x80), 0};
const   uint8_t TEMP[]  PROGMEM = {'t','e','m','p','e','r','a',('t' | 0x80)};
const   uint8_t STAT[]  PROGMEM = {'s','t','a','t','u',('s' | 0x80), 0, 0};
const   uint8_t VOLT[]  PROGMEM = {'v','o','l','t','a','g',('e' | 0x80), 0};
#define MAXCOMMAND 16

typedef enum {
	NOINT,				// no interrupt now
	TIMEINT,			// Timer overflow2 interrupt
	CHARINT,			// Regular character interrupt
	CMDINT,				// Command interrupt
	PININT				// Pin change interrupt
	} intrpttype;

/* Variables */

volatile intrpttype     interruptstatus;
volatile uint32_t       ticks;
volatile uint16_t       volts;
volatile uint16_t       v2volts;
volatile uint16_t       v3volts;
volatile uint16_t       volts_filt;
volatile uint16_t       v2volts_filt;
volatile uint16_t       v3volts_filt;
volatile uint16_t       highestV = 0;
volatile uint16_t       lowestV = 0x7fff;
volatile uint16_t       temp;
volatile uint16_t       extemp;
volatile uint16_t       temp_filt;
volatile uint16_t       extemp_filt;
volatile int8_t         buffer_one = 1;
         int16_t        buffer_address = 0;
         int16_t        page_address = 0;

        uint16_t        voltage_cal = VOLTSCALNUM;
        uint16_t        voltage_offset = VOLT_OFFSET;
        uint16_t        volts_max;
        uint16_t        volts_min;
        uint16_t        v2_cal = VOLTSCALNUM;
        uint16_t        v2_offset = VOLT_OFFSET;
        uint16_t        v3_cal = VOLTSCALNUM;
        uint16_t        v3_offset = VOLT_OFFSET;
        uint16_t        temperature_cal = TEMPCALNUM;
        uint16_t        temperature_offset = TEMP_OFFSET;
        uint16_t        extemp_cal = VOLTSCALNUM;
        uint16_t        extemp_offset = VOLT_OFFSET;
        uint16_t        max_temp;
        uint16_t        value_to_convert;
        uint8_t         status;
        uint16_t        calibration_value;
        int32_t         serial_number;
        uint8_t         bus_address;
        uint8_t         shunt_enable = 1;
//      int8_t          command_ready = 0;
        uint8_t         count, last_count;
        /* The reason for full_cycle is to make sure we have a complete
        * measurement before we check for min, max, and shunting voltages */
        uint8_t full_cycle = 0;

/*************************************************************
* increment a buffer pointer
*************************************************************/
uint8_t bump(uint8_t pointer)
{
    if (++pointer >= USART_BUFFER_SIZE)
        pointer = 0;
    return pointer;
}

/*************************************************************
* are there characters in a fifo
*************************************************************/
uint8_t DataInBuffer(const fifo * bf)
{
    return (bf->get != bf->put);
}
    
/*************************************************************
* is a whole string in a fifo?
*************************************************************/
uint8_t StringInBuffer(const fifo * bf)
{
	if (bf->get != bf->put)
	{
		uint8_t myget;
		for (myget = bf->get; bf->get != bf->put; myget = bump(myget))
		{
			if (bf->buffer[myget] == 0)
				return 1;
		}
	}
	return 0;
}   

/*************************************************************
* put a character into a fifo
*************************************************************/
uint8_t fifo_put(fifo *bf, uint8_t inchar)
{
    uint8_t myput = bump(bf->put);

    if (myput == bf->get)
        {
        return 0;
        }
    
    bf->buffer[bf->put] = inchar;
    bf->put = myput;
    return 1;
}

/*************************************************************
* get a character from a fifo
*************************************************************/
uint8_t fifo_get(fifo *bf, uint8_t *outchar)
{
    if (bf->put == bf->get)
        {
        return 0;
        }
        
    *outchar = bf->buffer[bf->get];
    bf->get = bump(bf->get);
    return 1;
}

/* Interrupt handlers */
void SIG_PIN_CHANGE0() __attribute__ ( ( signal ) );
void SIG_PIN_CHANGE0()
{
	// turn off pin change interrupts for port B
	PCIFR |=  (1 << PCIE0);
	PCICR &= ~(1 << PCIE0);
	interruptstatus = PININT;
}

void SIG_PIN_CHANGE2() __attribute__ ( ( signal ) );
void SIG_PIN_CHANGE2()
{
	// turn on usart
	UCSR0B |=  (( 1 << RXCIE0 ) | ( 1 << RXEN0 ) | ( 1 << TXEN0 ));
	// turn off pin change interrupts for port D
	PCIFR |=  (1 << PCIE2);
	PCICR &= ~(1 << PCIE2);
	interruptstatus = PININT;
}

void SIG_USART_RECV() __attribute__ ( ( signal ) );
void SIG_USART_RECV()
{
    uint8_t status;
    uint8_t data;
//    LED_PORT &= ~(1 << STATUS_LED);
    /* Read the received data */
    status = UCSR0A;
    data = UDR0;
    
	if ((status & (1 << FE0)) && (data == 0x00))
	{	// the bus seems disconnected...
		// turn off usart
    	UCSR0B &= ~(( 1 << RXCIE0 ) | ( 1 << RXEN0 ) | ( 1 << TXEN0 ));
    	// turn on pin change interrupts for port D
    	PCIFR |= (1 << PCIE2);
    	PCICR |= (1 << PCIE2);
	}
	else
	{
	    if (((data >= 0x20) && (data < 0x80)) || (data == 0x0D))
	    {
	        if (data == 0x0D)
	        {
	            fifo_put(&rxchars, 0);
	            interruptstatus = CMDINT;
//	            LED_PORT &= ~(1 << STATUS_LED);
	        }
	        else
	        {
	            fifo_put(&rxchars, data);
	            interruptstatus = CHARINT;
	        }
	    }
	}
//  LED_PORT |= (1 << STATUS_LED);
}

void SIG_USART_DATA() __attribute__ ( ( signal ) );
void SIG_USART_DATA()
{
    uint8_t tmp;

    /* Check if all data is transmitted */
    if (fifo_get(&txchars, &tmp))
    {
        UDR0 = tmp;  /* Start transmition */
    }
    else
    {
        UCSR0B &= ~(1<<UDRIE0);         /* Disable UDRE interrupt */
    }
}

#ifdef USETIMER1
void SIG_OVERFLOW1() __attribute__ ( ( signal ) );
void SIG_OVERFLOW1()                  /* signal handler for tcnt1 overflow interrupt */
#else
void SIG_OVERFLOW2() __attribute__ ( ( signal ) );
void SIG_OVERFLOW2()                  /* signal handler for tcnt1 overflow interrupt */
#endif
{
    // wake up the atod!
    ADCSRA = ADCFAKE;
    loop_until_bit_is_clear(ADCSRA, ADSC);
    
    // measure our values
//    LED_PORT &= ~(1 << STATUS_LED);
    volts   = filter(ATOD_read(BatteryV), &volts_filt);
    v2volts = filter(ATOD_read(V2), &v2volts_filt);
    v3volts = filter(ATOD_read(V3), &v3volts_filt);
    temp    = filter(ATOD_read(LM34_temp), &temp_filt);
    extemp  = filter(ATOD_read(EXT_temp), &extemp_filt);
    if (volts > highestV)
    {
        highestV = volts;
    }
    if (volts < lowestV)
    {
        lowestV = volts;
    }
//    LED_PORT |= (1 << STATUS_LED);
    
    interruptstatus = TIMEINT;
    ++ticks;
}

/* Initialize USART */
void USART_Init( uint16_t baudrate )
{
    /* Set the baud rate */
    UBRR0H = (uint8_t) (baudrate>>8);
    UBRR0L = (uint8_t) baudrate;

    /* Set frame format: 8 data 2stop */
    UCSR0C = (1<<USBS0)|(1<<UCSZ01)|(1<<UCSZ00);              //For devices with Extended IO

	// turn on usart
	UCSR0B |= ( 1 << TXEN0 );
    /* Flush receive buffer */
    rxchars.get = 0;
    rxchars.put = 0;
    txchars.get = 0;
    txchars.put = 0;
}


/* UART Read and write functions */
uint8_t uart_getchar( void )
{
    uint8_t tmp;
    while (!fifo_get(&rxchars, &tmp));  /* Wait for incomming data */

    return tmp;           /* Return data */
}

void uart_putchar( uint8_t data )
{
    if (!(UCSR0B & (1<<UDRIE0))) // we aren't sending now
    {   // send it
        UDR0 = data;
        UCSR0B |= (1<<UDRIE0);  
    }                   /* Enable UDRE interrupt */
    else
    {   // queue it
        // keep trying till it goes
        while (!fifo_put(&txchars, data));
    }
   if (data == '\n')
    {
        while (!fifo_put(&txchars, '\r'));
    }
}

void uart_putstring(char * fmt, ...)
{
    uint8_t outbuf[32];
    uint8_t i;
    va_list ap;

    va_start(ap, fmt);
    vsprintf(outbuf,fmt,ap);
    va_end(ap);

    for(i=0;i<32;i++) {
        if(outbuf[i] == 0)
            break;

        uart_putchar(outbuf[i]);
    }
}

uint16_t ATOD_read(uint8_t channel)
{
    ADMUX = (ADCREF << REFS0) | channel;
    ADCSRA = ADCSTART;
    loop_until_bit_is_clear(ADCSRA, ADSC);
    return ADC;
}

/**********************************************************************
* Filter a value
* inputs:
*  value to filter, must be less than 16383
*  last filter sum, must be less than 65535
* FILTER_CONST must be power of 2 between 2 and 64
**********************************************************************/
uint16_t filter(uint16_t value, uint16_t volatile * last)
{
    uint16_t result;
    
    *last = ((*last * (FILTER_CONST - 1)) / FILTER_CONST) + value;
    result = *last / FILTER_CONST;
    if ((*last / (FILTER_CONST - 1)) & 1)
    {
        result++;
    }
    return result;
}

/**********************************************************************
* convert a sample to a resulting value using the following formula:
* output_value = ((input_value * calibration) - offset) >> 10
**********************************************************************/
uint8_t convert_voltage(uint8_t* output, uint16_t value, uint16_t cal, int16_t off)
{
    int32_t value_to_convert = ((int32_t) value * (int32_t) cal) - off;
    uint16_t myval = (uint16_t) (value_to_convert >> 10);
    uint8_t cvtbuf[6];
    uint8_t *outptr = cvtbuf;
 
    if (value_to_convert & 0x200)
    {
        myval++;
    }
    
    vsprintf(cvtbuf,"%04dV",&myval);

    *output++ = *outptr++;
    *output++ = *outptr++;
    *output++ = '.';
    *output++ = *outptr++;
    *output++ = *outptr++;
    *output++ = *outptr++;
    *output = 0;
    return 1;
}

void convert_temperature(uint8_t* output, uint16_t value, uint16_t cal, int16_t off)
{
    int32_t value_to_convert = ((int32_t) value * (int32_t) cal) - off;
    uint16_t myval = (uint16_t) (value_to_convert >> 10);
 
    if (value_to_convert & 0x200)
    {
        myval++;
    }
    
    vsprintf(output,"%03dF",&myval);
}

/* get a command to evaluate */
/* void get_command(uint8_t * cmd_str, uint8_t length)
{
    uint8_t count = 0;
    
    do
    {
        *cmd_str = uart_getchar();
    } while ((*(cmd_str++) != 0x0d) && (count++ < (length - 1)));
    *--cmd_str = 0;
} */

uint8_t *get_string()
{
    // returns a string from the uart buffer
    uint8_t i;
    static uint8_t cmd_buffer[MAXCMDLEN];

    cmd_buffer[0] = 0;

    if(DataInBuffer(&rxchars))
        {
        for(i=0; i<MAXCMDLEN; i++) {
            cmd_buffer[i] = uart_getchar();
            if(cmd_buffer[i] == 0) {
                break;
            }
        }
    }

    return(&cmd_buffer[0]);

}


/******************************************************************************
*
* convert two characters into a decimal address, returning 0 if there is a 
* problem.
*
******************************************************************************/
uint8_t get_address(uint8_t ch1, uint8_t ch2)
{
    uint8_t temp = 0;
    
    if (isdigit(ch1) && isdigit(ch2))
    {
        temp = ((ch2 - '0') * 10) + ch1 - '0' ;
    }
    else
    {
        temp = 0xfe;
    }
    return temp;
}

/******************************************************************************
*
* check comand against list is program space
*  Commands can be shortened by ending with a period. Leaves
*  table_ptr pointing at first character of next entry.
*  Returns true if the entry matched.
*
******************************************************************************/
uint8_t check_command(prog_char *table_ptr)
{
    uint8_t retval = 1;
    uint8_t table_char = 0;
    uint8_t cmd_char;
    
    // first character can't be a '.'
    do 
    {
        cmd_char = *(command_pointer++);
        table_char = pgm_read_byte_near(table_ptr++);
        if ((cmd_char != '.') && (cmd_char != (table_char & 0x7f)))
        {
            retval = 0;
        }
    } while (retval && (cmd_char != '.') && !(table_char & 0x80));
    return retval;
}

/*****************************************************************************
* Get_Value takes a string, ignores anything that isn't a number, and converts
* it to a binary number
*****************************************************************************/
uint16_t get_value(uint8_t *instring)
{
    uint8_t mychar;
    uint16_t retval  = 0;
    
    do
    {
    mychar = *instring++;
        if ((mychar >= '0') && (mychar <= '9'))
        {
            retval = (retval * 10) + (mychar & 0x0f);
        }
    }
    while (mychar != 0);
    return retval;
}

/*****************************************************************************
*
* Evaluate and Execute a command from the EVILBUS
*
*****************************************************************************/
void evaluate_command(uint8_t * command)
{
    uint8_t cmd_num;
    uint8_t cmd_addr;
    prog_char * cmd_tab_pointer;
    uint8_t * cmd_ptr_save;
    uint8_t converted[16];
    uint16_t myvalue;
    
    command_pointer = command;
    cmd_addr = get_address(*(command_pointer++),*(command_pointer++));
    cmd_ptr_save = command_pointer;
    if (((cmd_addr == bus_address) || (cmd_addr == 99)) && (command[2] != '.'))
    {
        TCCR2B = 0;      // turn off timer 2
//      uart_putstring("\n command was:%s\n", command);
        LED_PORT &= ~(1 << STATUS_LED);

        cmd_tab_pointer = (prog_char *) VOLT;
        for (cmd_num = 0; cmd_num < MAXCOMMAND; cmd_num++)
        {
        
            if (check_command(cmd_tab_pointer))
            {
                switch (cmd_num)
                {
                    case 0:     // voltage
                        convert_voltage(converted, volts, voltage_cal, voltage_offset);
                        #ifdef LONG_RESPONSE
                        uart_putstring(" %02dV %s\n", bus_address, converted);
                        #else
                        uart_putstring(" %s\n", volts_converted);
                        #endif
                        break;
                    case 1:     // status
                        #ifdef LONG_RESPONSE
                        uart_putstring(" %02dS 0x%02x\n", bus_address, status);
                        #else
                        uart_putstring(" 0x%02x\n", status);
                        #endif
                        break;
                    case 2:     // temperature
                        convert_temperature(converted, temp, temperature_cal, temperature_offset);
                        #ifdef LONG_RESPONSE
                        uart_putstring(" %02dT %s\n", bus_address, converted);
                        #else
                        uart_putstring(" %s\n", converted);
                        #endif
                        break;
                    case 3:     // voltage 2
                        convert_voltage(converted, v2volts, v2_cal, v2_offset);
                        #ifdef LONG_RESPONSE
                        uart_putstring(" %02dV2 %s\n", bus_address, converted);
                        #else
                        uart_putstring(" %s\n", v2_string);
                        #endif
                        break;
                    case 4:     // voltage 3
                        convert_voltage(converted, v3volts, v3_cal, v3_offset);
                        #ifdef LONG_RESPONSE
                        uart_putstring(" %02dV3 %s\n", bus_address, converted);
                        #else
                        uart_putstring(" %s\n", v3_string);
                        #endif
                        break;
                    case 5:     // set high voltage limit
                        /************************************************************************
                        * Create a binary value from ascii number using the following formula:
                        * output_value = ((input_value << 10) + offset) / calibration
                        * rounding requires I go one more bit so I can tell if it needs it!
                        ************************************************************************/
                        myvalue = get_value(command_pointer);
                        myvalue = (uint16_t) (((uint32_t)((uint32_t)myvalue << 11) + voltage_offset) / voltage_cal);
                        // round it!
                        volts_max = (myvalue >> 1);
                        if (myvalue & 1)
                        {
                            volts_max++;
                        }
                        if (cmd_addr != 99)
                        {
                            #ifdef LONG_RESPONSE
                            uart_putstring(" %02dL %d\n", bus_address, volts_max);
                            #else
                            uart_putstring(" %d\n", volts_max);
                            #endif
                        }
                        eeprom_write_word((uint16_t*) HIGH_VOLTS_ADDR, volts_max);
                        break;
                    case 6:     // set low voltage limit
                        /************************************************************************
                        * Create a binary value from ascii number using the following formula:
                        * output_value = ((input_value << 10) + offset) / calibration
                        * rounding requires I go one more bit so I can tell if it needs it!
                        ************************************************************************/
                        myvalue = get_value(command_pointer);
                        myvalue = (uint16_t) (((uint32_t)((uint32_t)myvalue << 11) + voltage_offset) / voltage_cal);
                        // round it!
                        if (myvalue & 1)
                        {
                            volts_min = (myvalue >> 1) + 1;
                        }
                        else
                        {
                            volts_min = (myvalue >> 1);
                        }
                        if (cmd_addr != 99)
                        {
                            #ifdef LONG_RESPONSE
                            uart_putstring(" %02dL %d\n", bus_address, volts_min);
                            #else
                            uart_putstring(" %d\n", volts_min);
                            #endif
                        }
                        eeprom_write_word((uint16_t*) LOW_VOLTS_ADDR, volts_min);
                        break;
                    case 7:     // read min voltage
                        #ifdef LONG_RESPONSE
                        convert_voltage(converted, lowestV, voltage_cal, voltage_offset);
                        uart_putstring(" %02dI %s\n", bus_address, converted);
                        #else
                        uart_putstring(" %s\n", converted);
                        #endif
                        break;
                    case 8:     // read max voltage
                        convert_voltage(converted, highestV, voltage_cal, voltage_offset);
                        #ifdef LONG_RESPONSE
                        uart_putstring(" %02dA %s\n", bus_address, converted);
                        #else
                        uart_putstring(" %s\n", converted);
                        #endif
                        break; 
                    case 9:     // clear min/max history
                        highestV = 0;
                        lowestV = 0x7fff;
                        status &= ~BEENLOW_STATUS;
                        LED_PORT |= (1 << UNDERVOLT_LED);
                        break;
                    case 10:        // disable shunt
                        status |= DISABLED_STATUS;
                        break;
                    case 11:        // enable shunt
                        status &= ~DISABLED_STATUS;
                        break;
                    case 12:        // store eeprom value
                        uart_putstring(" Not Implemented!\n");
                        break;
                    case 13:        // read high voltage limit setting
                        convert_voltage(converted, volts_max, voltage_cal, voltage_offset);
                        #ifdef LONG_RESPONSE
                        uart_putstring(" %02dH %s\n", bus_address, converted);
                        #else
                        uart_putstring(" %s\n", converted);
                        #endif
                        break;
                    case 14:        // read low voltage limit setting
                        convert_voltage(converted, volts_min, voltage_cal, voltage_offset);
                        #ifdef LONG_RESPONSE
                        uart_putstring(" %02dL %s\n", bus_address, converted);
                        #else
                        uart_putstring(" %s\n", converted);
                        #endif
                        break;
                    case 15:        // read external temperature
                        convert_temperature(converted, extemp, temperature_cal, temperature_offset);
                        #ifdef LONG_RESPONSE
                        uart_putstring(" %02dX %s\n", bus_address, converted);
                        #else
                        uart_putstring(" %s\n", converted);
                        #endif
                        break;
                    default:
                        uart_putstring(" ?\n");
                        break;
                }
                break;
            }
            cmd_tab_pointer += 8;
            command_pointer = cmd_ptr_save;
        }
        TCCR2B = TCCR2_DIV | ((TCCR2_MODE & 0x04) << 1);      // turn on timer 2
    }
}

/***********************************************************************
*
* Calibrate Internal R/C Oscillator
*
* first synchronize with the 32768 hz tc2, then count 8mhz tc1 clocks
* it takes to do 4 more tc2 ticks.
*
***********************************************************************/
void cal_oscillator(void)
{
    uint8_t temp;
    uint16_t    first_count, this_count, next_count;

	#define CALWAIT 20
	
    temp = TCNT2 + 2;
    if (temp < (MAX_OSC_COUNT - 2 - CALWAIT))
	{
	    while (temp != TCNT2);
	    first_count = TCNT1;
	    temp = temp + CALWAIT;
	    while (temp != TCNT2);
	    next_count = TCNT1;
	    this_count = next_count - first_count;
	    if (this_count > MAX_OSC_COUNT)
	    {
	        OSCCAL -= 1;
	    }
	    else if (this_count < MIN_OSC_COUNT)
	    {
	        OSCCAL += 1;
	    }
	}
}

int main(void)
{
    uint8_t *cmd_ptr;
    uint8_t eval_time, start_time, max_eval = 0;

    USART_Init(BAUDRATEDIV);

    // get timer 2 running so it can stabilize
    // Async & CTC Modes
    TCNT0 = 0;
    ASSR = 1 << AS2;        // set asych mode
    OCR2A = OCR2_COUNT;
//    OCR2A = 127;
    TCCR2A = (TCCR2_MODE & 0x03);
    TCCR2B = TCCR2_DIV | ((TCCR2_MODE & 0x04) << 1);      // turn on timer 2

    while (ASSR != (1 << AS2));     // wait till it is done
    //                        ^ do nothing
    
    // get the reference up and running
    ADMUX = (ADCREF << REFS0);

    // Turn off things we don't need like:
    // two wire interface, timer 0, and SPI
    PRR = (1 << PRTWI) | (1 << PRTIM0);
    // and the digital inputs on analog pins
    DIDR0 = ANALOG_PINS;
    
    // start the 16 bit counter ticking
    TCCR1A = TCCR1_MODE & 0x03;
    TCCR1B = ((TCCR1_MODE << 1) & 0x18) | TCCR1_DIV;

#ifdef USETIMER1
    ICR1 = ICR1_COUNT;
    TIFR1 = 1 << TOV1;       // clear the flag
    TIMSK1 = 1 << TOIE1;     // enable T1 overflow interrupts
#else
    TIFR2 = 1 << TOV2;
    TIMSK2 = 1 << TOIE2;
#endif

#ifdef DEBUG
	TCCR0B = 5;				// divide by 1024
#endif

    voltage_offset = eeprom_read_word((uint16_t*) VOLTAGE_OFFSET_ADDR);
    voltage_cal = eeprom_read_word((uint16_t*) VOLTAGE_CAL_ADDR);
    temperature_offset = eeprom_read_word((uint16_t*) TEMPERATURE_OFFSET_ADDR);
    temperature_cal = eeprom_read_word((uint16_t*) TEMPERATURE_CAL_ADDR);
    v2_offset = eeprom_read_word((uint16_t*) V2_OFFSET_ADDR);
    v2_cal = eeprom_read_word((uint16_t*) V2_CAL_ADDR);
    v3_offset = eeprom_read_word((uint16_t*) V3_OFFSET_ADDR);
    v3_cal = eeprom_read_word((uint16_t*) V3_CAL_ADDR);
    volts_min = eeprom_read_word((uint16_t*) LOW_VOLTS_ADDR);
    volts_max = eeprom_read_word((uint16_t*) HIGH_VOLTS_ADDR);
    max_temp = eeprom_read_word((uint16_t*) TEMP_HOT_ADDR);
    eeprom_read_block(&serial_number, (uint8_t*) SERIAL_NUMBER_ADDR, 4);
    OSCCAL = eeprom_read_byte((uint8_t*)OSCCAL_BYTE_ADDR);

    bus_address = eeprom_read_byte((uint8_t*) BUS_UNIT_ADDR);
    sei();                  // enable interrupts

	// set which pins to use the pin change interrupts for
	PCMSK2 |= (1 << PCINT16);
	PCMSK0 |= (1 << PCINT0);
	// turn on pin change interrupts for port D
	PCIFR |= (1 << PCIE2);
	PCICR |= (1 << PCIE2);

    // Make output pins output, and turn them off!!
    LED_PORT &= ~((1 << RED_LED) | (1 << YELLOW_LED) | (1 << GREEN_LED) | (1 << STATUS_LED));
    LED_DDR |= ((1 << RED_LED) | (1 << YELLOW_LED) | (1 << GREEN_LED) | (1 << STATUS_LED));
    OPTO_PORT |= ((1 << REG_VOLTS_HI) | (1 << REG_TOO_MUCH));
    OPTO_DDR |= ((1 << REG_VOLTS_HI) | (1 << REG_TOO_MUCH));

    // and the FET port
    FET_PORT &= ~(1 << FET_gate);
    FET_DDR |= (1 << FET_gate);

    // wait for 1 sample period for the reference to come up
    while (ticks <= BLINKDELAY);
    ticks = 0;

    uart_putstring(SIGNON_STRING);
    uart_putstring(VERSION_STRING);
    // the LED's start on, then blink red, yellow, green before we start
    LED_PORT |= ((1 << RED_LED) | (1 << YELLOW_LED) | (1 << GREEN_LED));
    LED_PORT &= ~(1 << RED_LED);
    while (ticks <= BLINKDELAY);
    ticks = 0;
    LED_PORT |= ((1 << RED_LED) | (1 << YELLOW_LED) | (1 << GREEN_LED));
    LED_PORT &= ~(1 << YELLOW_LED);
    while (ticks <= BLINKDELAY);
    ticks = 0;
    LED_PORT |= ((1 << RED_LED) | (1 << YELLOW_LED) | (1 << GREEN_LED));
    LED_PORT &= ~(1 << GREEN_LED);

    uart_putstring("%02d S/N: %d\n",bus_address, serial_number);
    while (ticks <= BLINKDELAY);
    ticks = 0;
    LED_PORT |= ((1 << RED_LED) | (1 << YELLOW_LED) | (1 << GREEN_LED));
    
    volts_filt   = ATOD_read(BatteryV) * FILTER_CONST;
    v2volts_filt = ATOD_read(V2) * FILTER_CONST;
    v3volts_filt = ATOD_read(V3) * FILTER_CONST;
    temp_filt    = ATOD_read(LM34_temp) * FILTER_CONST;
    extemp_filt  = ATOD_read(EXT_temp) * FILTER_CONST;
    
    status |= DISABLED_STATUS;

    for(;;)
    {
//      LED_PORT |= (1 << STATUS_LED);
        switch (interruptstatus)
        {
        	case CMDINT:
		        if (StringInBuffer(&rxchars))
		        {
					#ifdef DEBUG
						start_time = TCNT0;
					#endif
			        if (strlen(cmd_ptr = get_string()) >= 4)
			        {
		
			            evaluate_command(cmd_ptr);
			            LED_PORT |= (1 << STATUS_LED);
			            // we did a command so we can't trust this measurement
			        }
			        #ifdef DEBUG
			        	eval_time = TCNT0 - start_time;
			        	max_eval = eval_time > max_eval ? eval_time : max_eval;
			        #endif
			    }
			    break;
        
			case TIMEINT:
		        // Check Over Voltage
		        if (volts > volts_max)
		        {
		            // turn off the red led
		            LED_PORT |= (1 << UNDERVOLT_LED);
		            // turn on the green led and the too much
		            LED_PORT &= ~(1 << OVERVOLT_LED);
		            OPTO_PORT &= ~(1 << REG_TOO_MUCH);
		            if (!(status & DISABLED_STATUS))
		            {
		                #ifdef ALLOW_SHUNT
		                FET_PORT |= (1 << FET_gate);
		                #endif
		            }
		            status |= (OVERVOLT_STATUS | SHUNTING_STATUS);
		            status &= ~BEENLOW_STATUS;
		        }
		        else
		        {
		            // turn off the green led and the too much
		            LED_PORT |= (1 << OVERVOLT_LED);
		            status &= ~(OVERVOLT_STATUS | SHUNTING_STATUS);
		            if (!(status & OVERVOLT_STATUS) && !(status & UNDERVOLT_STATUS) && !(status & OVERVOLT_STATUS)) 
		                {
		                OPTO_PORT |= 1 << REG_TOO_MUCH;
		                }
		            FET_PORT &= ~(1 << FET_gate);
		        }
		        
		        // Check Under Voltage
		        if (volts < volts_min)
		        {
		            // turn on the red and yellow LEDs
		            LED_PORT &= ~((1 << UNDERVOLT_LED) | (1 << UNDREQUAL_LED));
		            // turn on the reg too_much
		            OPTO_PORT &= ~(1 << REG_TOO_MUCH);
		            status |= UNDERVOLT_STATUS | BEENLOW_STATUS;
		        }
		        else
		        {
		            // turn off the yellow LED
		            LED_PORT |= (1 << UNDREQUAL_LED);
		            status &= ~UNDERVOLT_STATUS;
		            if (!(status & OVERVOLT_STATUS) && !(status & UNDERVOLT_STATUS) && !(status & OVERVOLT_STATUS)) 
		                {
		                OPTO_PORT |= 1 << REG_TOO_MUCH;
		                }
		        }
		        
		        // Check for Over Temp
		        if (temp > max_temp)
		        {
		            // turn on the reg too_much
		            OPTO_PORT &= ~(1 << REG_TOO_MUCH);
		            status |= OVERTEMP_STATUS;
		            // and turn off the FET
		            FET_PORT &= ~(1 << FET_gate);
		        }
		        else
		        {
		            // turn off the reg too_much
		            status &= ~OVERTEMP_STATUS;
		            if (!(status & OVERVOLT_STATUS) && !(status & UNDERVOLT_STATUS) && !(status & OVERVOLT_STATUS)) 
		                {
		                OPTO_PORT |= 1 << REG_TOO_MUCH;
		                }
		        }
			        
			    // Should we go to sleep hard?
			    if (!(CHARGER_PIN & (1 << CHARGER_ON)) && !(UCSR0B & (1 << RXCIE0 )))
			    	{ // YES!
				    uint8_t temp = LED_PORT;
					// turn off usart
			    	UCSR0B &= ~(( 1 << RXCIE0 ) | ( 1 << RXEN0 ) | ( 1 << TXEN0 ));
			    	// turn on pin change interrupts for port B & D
			    	PCIFR |= (1 << PCIE2) | (1 << PCIE0);
			    	PCICR |= (1 << PCIE2) | (1 << PCIE0);
			    	// turn off the led's
			    	LED_PORT |= (1 << RED_LED) | (1 << YELLOW_LED) | (1 << GREEN_LED) | (1 << STATUS_LED);
		            set_sleep_mode ( SLEEP_MODE_PWR_DOWN );
		            sleep_mode ();
		            LED_PORT = temp;
			    	}
		        // we calibrate the oscillator every so often
		        if ((ticks & CALIBRATION_MASK) == 0)
		        {
//		            cal_oscillator();
		        }
	            break;
        	case CHARINT:
        	default:
		        interruptstatus = 0;
	            set_sleep_mode ( SLEEP_MODE_IDLE );
	            sleep_mode ();
	            break;
        }
    }
}
